# -*- python -*-
# $RCSfile: utils.py,v $
# $Revision: 1.44 $
# $Author: langer $
# $Date: 2005/05/13 18:43:46 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import lock
from oof2.common import debug
import string
import sys
import types

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to make a tuple out of its inputs.  This is useful
# because sometimes we want to use the same loop to iterate over
# things, even if the things are not in an indexable object.
#   This function makes no effort to tupify sublists or other
# internal components of the passed-in object.
def tupify(x):
    try:
        return tuple(x)
    except (TypeError, AttributeError):
        return (x,)
    
def listify(x):
    try:
        return list(x)
    except (TypeError, AttributeError):
        return [x]

# Argument is presumptively either a string or a list of strings.
# This is needed to pass around the profiles for the boundary
# conditions.
def string_listify(x):
    if type(x) == type(''):
        return [x]
    else:
        return x

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to return adjacent pairs of a list, i.e. given [1,2,3],
# it returns [(1,2), (2,3)].
def list_pairs(lst):
    return [ (lst[i], lst[i+1]) for i in range(len(lst)-1) ]

# Possible to write in terms of generators, if you're sure that
# nobody's going to try to use Python 2.1 or earlier.
# from __future__ import generators
# def list_pairs(lst):
#     for i in range(len(lst)-1):
#         yield (lst[i],lst[i+1])

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
# Function to return 1 if the argument is indexable and 0 otherwise
def isIndexable(x):
    try:
        len(x) # I suppose one could imagine python optimizing this away!
        return 1
    except TypeError:
        return 0
    except AttributeError:
        return 0

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
# Used by boundary-condition consistency check to see if small tuples
# overlap.  Returns a possibly-empty tuple of the intersecting
# components.
def tupleIntersect(x,y):
    result = ()  # Empty tuple.
    for xvalue in x:
        for yvalue in y:
            if xvalue == yvalue:
                result+=(xvalue,)
    return result

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Function to take a list of lists and flatten it into a single list.
def flatten(lols):
    result = []
    for l in lols:
        result += l
    return result

# Unfortunately, this version is much slower:
##   def flatten(lols): return [x for y in lols for x in y]

def flatten_all(lols):
    result = []
    for l in lols:
        if type(l) is types.ListType or type(l) is types.TupleType:
            result += flatten_all(l)
        else:
            result.append(l)
    return result

# Given a template which is a list of lists of lists to arbitrary
# depth, convert the flat datalist to a nested list with the same
# structure.  Each sublist of the result will have the same length as
# the corresponding sublist in the template.
def unflatten(template, datalist):
    which = 0
    result, which = _do_unflatten(template, datalist, which)
    return result

def _do_unflatten(template, datalist, which):
    result = []
    for obj in template:
        if type(obj) == types.ListType or type(obj) == types.TupleType:
            sublist, which = _do_unflatten(obj, datalist, which)
            result.append(sublist)
        else:
            result.append(datalist[which])
            which += 1
    return result, which

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Get a list of the classes to which an object or class belongs.
def classes(c):
    if type(c) == types.InstanceType:
        return classes(c.__class__)
    if type(c) == types.ClassType:
        if not c.__bases__:
            return [c]
        return [c] +  flatten(map(classes, c.__bases__))
    return []                           # not a class or an instance of a class

# Get the *names* of the classes to which an object or class belongs.
# This loses namespace information, so it's not as robust as using
# classes().  But it's more readable.
def classnames(c):
    return [cl.__name__ for cl in classes(c)]

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

def canonicalorder(p0, p1):
    if p0 < p1:
        return (p0, p1)
    return (p1, p0)

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Evaluate expressions, statements and files in the main oof namespace.

mainmodule = sys.modules['__main__']

def OOFexec(command):
    exec(command, mainmodule.__dict__)

def OOFeval(expr):
    return eval(expr, mainmodule.__dict__)

def OOFexecfile(file):
    execfile(file, mainmodule.__dict__)

# Define an existing object in the main oof namespace.
def OOFdefine(name, obj):
    mainmodule.__dict__[name] = obj

# Safe, restricted evaluation: can't import or do anything malicious.
# Can only be used to evaluate things defined in the main namespace.
# Is used when loading data files.

def OOFeval_r(expr):
    return mainmodule.__dict__[expr]

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Parse an argument list, e.g. from inside an "eval" when the
# arguments arrive in string form.

    
def argback(*args,**kwargs):
    return {"tuple":args, "dictionary":kwargs}

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Construct a name, based on the given name arg, that does not occur
# in the given list of other names.  If the given name isn't in the
# list, return it.  If it is, append "<number>" to it, where 'number'
# is chosen so that the result is unique.

# The list of names passed to uniqueName is almost always a global
# list, and so in a threaded environment we need to prevent
# simultaneous uniqueName calls with the same name list.  Since the
# calls are quick and not done in time sensitive situations, we put a
# single mutex lock inside uniqueName, instead of requiring all
# callers to implement their own lock.

import re

_uniqueNameLock = lock.Lock()

def uniqueName(name, othernames, exclude=None):
    _uniqueNameLock.acquire()
    others = othernames[:]
    try:
        if exclude is not None:
            try:
                others.remove(exclude)
            except ValueError:
                pass
        if name not in others:
            return name
        # Strip '<number>' suffix, if any.
        basename = re.split('<[0-9]+>$', name)[0]
        # Find any existing names of the form 'basename<number>'.
        expr = re.compile("^" + basename + "<[0-9]+>$")
        matches = filter(expr.match, others)
        if matches:
            # Find largest existing "<number>".
            suffixes = [x[len(basename)+1:-1] for x in matches]
            lastsuffix = max(map(eval, suffixes))
        else:
            lastsuffix = 1
        return "%s<%d>" % (basename, lastsuffix+1)
    finally:
        _uniqueNameLock.release()
        
# Special version that uses underscores, for menu names.
def menUniqueName(name, othernames):
    _uniqueNameLock.acquire()
    try:
        if name not in othernames:
            return name
        # Strip '_number' suffix, if any.
        basename = re.split('_[0-9]+$', name)[0]
        # Find any existing names of the form 'basename_number'.
        expr = re.compile("^" + basename + "_[0-9]+$")
        matches = filter(expr.match, othernames)
        if matches:
            # Find largest existing "number".
            suffixes = [x[len(basename)+1:] for x in matches]
            lastsuffix = max(map(eval, suffixes))
        else:
            lastsuffix = 1
        return "%s_%d" % (basename, lastsuffix+1)
    finally:
        _uniqueNameLock.release()

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

# Convert a menu item name to the form displayed by the GUI.
# Underscores are replaced by spaces, and double underscores are
# replaced by underscores.  The sleazy implementation here also
# converts double spaces to underscores, but nobody in his right mind
# would put double spaces in menu item names, right?

def underscore2space(name):
    return name.replace("_", " ").replace("  ", "_")

# And the not-quite-inverse.
def space2underscore(name):
    return name.replace(" ", "_")

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#



# Utility class for a list with no duplicates.  Defers uniquification
# until a request for a list member is made.  Provides a full suite of
# list emulation functions.
class UniqueList:
    def __init__(self, list=[]):
        self.data = []
        self.unique = 1 # Default case for empty list, gets zero'd in append.
        for l in list:
            self.append(l)

    def append(self, item):
        self.data.append(item)
        self.unique = 0


    # Uniquification function -- actual uniquification is deferred
    # until retrieval, so there is no overhead for repeated additions.
    def uniquify(self):
        if self.unique==0:
            self.data.sort()
            nuniq = 1
            for i in range(1, len(self.data)):
                if self.data[i] != self.data[i-1]:
                    if nuniq < i:
                        self.data[nuniq] = self.data[i]
                    nuniq += 1
            del self.data[nuniq:]
            self.unique = 1
            
    def remove(self, item):
        self.uniquify()
        try:
            i = self.data.index(item)
        except ValueError:
            pass # Removal of non-entry is harmless, unlike regular lists.
        else:
            del self.data[i]

    def index(self, value):
        self.uniquify()
        return self.data.index(value)

    # "Add" will only work if "other" is a sequence type.
    # Returns a new object.
    def __add__(self, other):
        try:
            len(other)
        except TypeError:
            raise TypeError(
                "Only sequence types can be appended to UniqueList.")
        else:
            self.uniquify()
            res = UniqueList()
            res.data = self.data[:]
            for o in other:
                res.append(o)
            return res

    # Allow UniqueLists to be appended to regular lists -- in
    # such a case, copy out the data references.
    def __radd__(self, other):
        self.uniquify()
        if type(other)==types.ListType:
            return other + self.data[:]
        raise TypeError("UniqueLists can only be appended to ordinary lists.")
    
    # Insertion is actually meaningless, since UniqueLists get sorted.
    # It's provided for completeness, discards "pos" and calls append.
    def insert(self, pos, value):
        self.append(value)

    # Remaining list emulation functions, except for "__len__", which
    # is in the parent class.
    def __getitem__(self, index):
        self.uniquify()
        return self.data[index]

    def __getslice__(self, start, end):
        self.uniquify()
        return self.data[start:end]
    
    def __setitem__(self, index, value):
        self.uniquify()
        self.data[index] = value

    def __delitem__(self, index):
        self.uniqiufy()
        del self.data[index]

    def __contains__(self, value):
        return value in self.data

    def __len__(self):
        self.uniquify()
        return len(self.data)

    def __repr__(self):
        self.uniquify()
        return "UniqueList("+`self.data`+")"


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

def weedlist(condition, liszt):
    nextspot = 0
    for x in liszt:
        if condition(x):
            if liszt[nextspot] is not x:
                liszt[nextspot] = x
            nextspot += 1
    del liszt[nextspot:]
    return liszt

#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#
    
def screenwidth():
    try:
        import curses
        scr = curses.initscr()
        width = scr.getmaxyx()[1]
        curses.endwin()
    except ImportError:
        width = 80
    return width

# Function to format a long string so that it fits into strings of
# length width.  Returns a list of strings guaranteed to be shorter
# than width, provided there are some spaces in there somewhere.
def format(line, width):
    linelist = [string.strip(ell) for ell in line.split("\n")]
    outlist = []
    for str in linelist:
        while len(str) > width:
            breakpoint = string.rfind(str," ",0,width)
            outlist.append(str[0:breakpoint])
            str = str[breakpoint+1:]
        else:
            outlist.append(str)
    return outlist


#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#=*=#

if __name__=='__main__':
    l0 = [0, [1,2],[3,[4]], 5]
    l1 = ['zero', 'one', 'two', 'three', 'four', 'five']
    print unflatten(l0, l1)
