# -*- python -*-
# $RCSfile: meshIO.py,v $
# $Revision: 1.30 $
# $Author: reida $
# $Date: 2005/06/16 14:07:56 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Menu commands for reading a Mesh from a file.  To be precise, a Mesh
# is re-created from a read Skeleton, and then the Fields, Equations,
# BoundaryConditions, etc. are read.

from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import masterelement
from oof2.common import debug
from oof2.common import labeltree
from oof2.common import utils
from oof2.common.IO import mainmenu
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.engine import bdycondition
from oof2.engine import fieldinit
from oof2.engine import materialmanager
from oof2.engine import meshcrosssection
from oof2.engine import skeletonboundary
from oof2.engine import skeletoncontext
from oof2.engine import skeletonelement
from oof2.engine.IO import materialmenu
from oof2.engine.IO import meshparameters
from oof2.engine.IO import profileIO
import oof2.SWIG.engine.equation
import oof2.SWIG.engine.field
import oof2.SWIG.engine.material
import oof2.engine.mesh

OOFMenuItem = oofmenu.OOFMenuItem

OOF = mainmenu.OOF

meshmenu = OOF.LoadData.addItem(OOFMenuItem('Mesh'))

def writeMesh(datafile, meshcontext, skelcontext):
    skelpath = skelcontext.path()
    
    # Create mesh.
    datafile.startCmd(meshmenu.New)
    datafile.argument('name', meshcontext.name())
    masterelems = [`ename` for ename in meshcontext.elementdict.values()]
    datafile.argument('masterelems', masterelems)
    datafile.argument('skeleton', skelpath)
    datafile.endCmd()

    femesh = meshcontext.femesh()
    # Fields
    definedfields= []
    activefields= []
    inplanefields= []
    for field in femesh.all_fields():
        definedfields.append(field.name())
        if femesh.is_active_field(field):
            activefields.append(field.name())
        if femesh.in_plane(field):
            inplanefields.append(field.name())
    if definedfields:
        datafile.startCmd(meshmenu.Field)
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('defined', definedfields)
        datafile.argument('active', activefields)
        datafile.argument('inplane', inplanefields)
        datafile.endCmd()

    # Explicitly assigned materials
    explicitMats = {}
    for element in femesh.element_iterator():
        mat = element.getExplicitMaterial()
        if mat:
            try:
                elementList = explicitMats[mat]
            except KeyError:
                explicitMats[mat] = elementList = []
            elementList.append(element.get_index())

    # Make sure that the explicitly assigned Materials are defined in
    # the data file.  If they don't appear in the Microstructure, they
    # haven't been saved in the data file.
    existingMats = oof2.SWIG.engine.material.getMaterials(
        meshcontext.getMicrostructure())
    missingMats = [mat for mat in explicitMats.keys()
                   if mat not in existingMats]
    if missingMats:
        # Construct a list of Properties already defined in the data
        # file, so that they're not written twice.
        excludeProps = {}
        for mat in existingMats:
            for prop in mat.properties():
                excludeProps[prop.registration().name()] = prop
        # Write the missing materials to the data file.
        materialmenu.writeMaterials(datafile, missingMats, excludeProps)
    
    for mat, elementList in explicitMats.items():
        datafile.startCmd(meshmenu.AssignMaterial)
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('elements', elementList)
        datafile.argument('material', mat.name())
        datafile.endCmd()

    # Field initializers
    for field in femesh.all_fields():
        init = meshcontext.get_initializer(field)
        if init:
            datafile.startCmd(meshmenu.Initialize_Field)
            datafile.argument('mesh', meshcontext.path())
            datafile.argument('field', field)
            datafile.argument('initializer', init)
            datafile.endCmd()

    # Equations
    equations = [eqn.name() for eqn in femesh.all_equations()]
    if equations:
        datafile.startCmd(meshmenu.Equation)    
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('equations', equations)
        datafile.endCmd()

    # Named profiles, if required.
    for bcname in meshcontext.allBndyCondNames():
        bc = meshcontext.getBdyCondition(bcname)
        if bc.profile.is_named():
            profileIO.writeprofile(datafile, bc.profile.name, bc.profile.obj)
        
    # Boundary conditions
    for bcname in meshcontext.allBndyCondNames():
        bc = meshcontext.getBdyCondition(bcname)
        datafile.startCmd(meshmenu.Boundary_Condition)    
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('bcname', bcname)
        datafile.argument('bc', bc)
        datafile.endCmd()

    # Field values
    # fields = ["Displacement", "Temperature"] : ListOfStrings
    # field_values = [(node_index, fvalue0, fvalue1, fvalue2), ....]
    # : ListOfTuplesOfIntFloats

    fields = femesh.all_fields() + \
             [field.out_of_plane() for field in femesh.all_fields()
              if not field.in_plane(femesh)]
    if fields:
        fvalues = []
        for node in femesh.funcnode_iterator():
            fv = [node.index()]
            for field in fields:
                for i in range(field.ndof()):
                    fv.append(field.dof(femesh, node, i).value())
            fvalues.append(tuple(fv))
        datafile.startCmd(meshmenu.Load_Field)    
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('fields', [field.name() for field in fields])
        datafile.argument('field_values', fvalues)
        datafile.endCmd()

    # Cross sections:
    for csname in meshcontext.cross_sections.all_names():
        cs = meshcontext.cross_sections[csname]
        datafile.startCmd(meshmenu.CrossSection)
        datafile.argument('mesh', meshcontext.path())
        datafile.argument('name', csname)
        datafile.argument('cs', cs)
        datafile.endCmd()


def getMyMasterElementDict(masterelems):
    edict = {}
    masterdict = masterelement.getMasterElementDict()
    for ename in masterelems:
        try:
            master = masterdict[ename]
            edict[master.nsides()] = master
        except KeyError:
            raise ooferror.ErrUserError("Element type \"%s\" is unknown."
                                        % ename)
    return  edict

## TODO: This probably shouldn't hard-code the realmaterial material
## factory function, although at the moment there's not much chance of
## ever wanting to store Meshes that use other factory functions.

def _newMesh(menuitem, name, masterelems, skeleton):
    skel = skeletoncontext.skeletonContexts[skeleton].getObject()
    edict = getMyMasterElementDict(masterelems)
    femesh = skel.femesh(edict, skeletonelement.SkeletonElement.realmaterial)
    oof2.engine.mesh.meshes.add(labeltree.makePath(skeleton)+[name], femesh,
                           skeleton=skel, elementdict=edict,
                           materialfactory=skeletonelement.SkeletonElement.realmaterial)
    
meshmenu.addItem(OOFMenuItem(
    'New',
    callback=_newMesh,
    params=[parameter.StringParameter('name', tip="Name for the Mesh."),
            parameter.ListOfStringsParameter('masterelems',
                                             tip="Names of Master Elements."),
            whoville.WhoParameter('skeleton',
                                  skeletoncontext.skeletonContexts,
                                  tip=parameter.emptyTipString)
            ],
    help="Load a Mesh. Used internally in Mesh data files.",
    discussion="<para>Load a &mesh; from a datafile.</para>"
    ))


getFieldObj = oof2.SWIG.engine.field.getCompoundField

def _meshFields(menuitem, mesh, defined, active, inplane):
    themesh = oof2.engine.mesh.meshes[mesh].femesh()
    
    for fname in defined:
        field = getFieldObj(fname)
        themesh.define_field(field)
        switchboard.notify("field defined", mesh, field.name(), 1)

    for fname in active:
        field = getFieldObj(fname)
        themesh.activate_field(field)
        switchboard.notify("field activated", mesh, field.name(), 1)

    for fname in inplane:
        field = getFieldObj(fname)
        field.set_in_plane(themesh, 1)
        switchboard.notify("field inplane", mesh, field.name(), 1)

meshmenu.addItem(OOFMenuItem(
    'Field',
    callback=_meshFields,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.ListOfStringsParameter('defined',
                                             tip="Defined Fields."),
            parameter.ListOfStringsParameter('active', tip="Active Fields."),
            parameter.ListOfStringsParameter('inplane', tip="In-plane Fields.")],
    help="Load Fields. Used internally in Mesh data files.",
    discussion="""<para>
    Load the list of defined &fields; from a saved &mesh;. 
    </para>"""
    ))

getEquationObj = oof2.SWIG.engine.equation.getEquation

def _meshEquations(menuitem, mesh, equations):
    themesh = oof2.engine.mesh.meshes[mesh].femesh()

    for eqn in equations:
        equation = getEquationObj(eqn)
        themesh.activate_equation(equation)
        switchboard.notify('equation activated', mesh, equation.name(), 1)

meshmenu.addItem(OOFMenuItem(
    'Equation',
    callback=_meshEquations,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.ListOfStringsParameter('equations',
                                             tip="Names of Equations.")],
    help="Load Equations. Used internally in Mesh data files.",
    discussion="""<para>
    Load the list of active &equations; from a saved &mesh;.
    </para>"""
    ))

def _meshBoundary_Conditions(menuitem, mesh, bcname, bc):
    bc.add_to_mesh(bcname, mesh)

meshmenu.addItem(OOFMenuItem(
    'Boundary_Condition',
    callback = _meshBoundary_Conditions,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.StringParameter('bcname',
                                      tip="Name of the boundary condition."),
            parameter.RegisteredParameter('bc', bdycondition.BC,
                                          tip=parameter.emptyTipString)],
    help="Load a boundary conditon. Used internally in Mesh data files.",
    discussion="""<para>
    Load a <link linkend='Section:Tasks:BoundaryCondition'>boundary
    condition</link> from a saved &mesh;.
    </para>"""
    ))

def _loadFieldValues(menuitem, mesh, fields, field_values):
    femesh = oof2.engine.mesh.meshes[mesh].femesh()
    for fv in field_values:
        index = fv[0]
        node = femesh.getNode(index)
        pointer = 1
        for fld in fields:
            field = getFieldObj(fld)
            for i in range(field.ndof()):
                field.setvalue(femesh, node, i, fv[pointer])
                pointer += 1

meshmenu.addItem(OOFMenuItem(
    'Load_Field',
    callback = _loadFieldValues,
    params=[whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                                  tip=parameter.emptyTipString),
            parameter.ListOfStringsParameter('fields',
                                             tip="Names of Fields."),
            parameter.ListOfTuplesOfIntFloatsParameter('field_values',
                                                       tip="Values of Fields.")],
    help="Load Fields Used internally in Mesh data files.",
    discussion="""<para>
    Load values for a &field; at the &nodes; of a saved &mesh;.  The
    <varname>fields</varname> parameter is a <link
    linkend='Object:list'>list</link> of the names of the &fields;
    that are defined on the &mesh;.  The
    <varname>field_values</varname> parameter is a list of tuples
    (Python lists with parentheses instead of square brackets).  Each
    tuple contains an integer node number, followed by the components
    of the listed fields in the listed order.
    </para>"""
    ))


def _initField(menuitem, mesh, field, initializer):
    # This routine is repeated almost verbatim in meshmenu.py.
    meshcontext = oof2.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        meshcontext.initialize_field(field, initializer)
        switchboard.notify("field initialized")
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    meshcontext.changed()


meshmenu.addItem(OOFMenuItem(
    'Initialize_Field',
    callback=_initField,
    params=[
    whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    meshparameters.FieldParameter('field', tip=parameter.emptyTipString),
    fieldinit.FieldInitParameter('initializer', tip=parameter.emptyTipString)
    ],
    help="Initialize a Field. Used internally in Mesh data files.",
    discussion="""<para>
    Initialize a &field; on a saved &mesh;.  If <link
    linkend='MenuItem:OOF.LoadData.Mesh.Load_Field'>field data</link>
    is saved as well, it will overwrite the &fields; from the
    initializer.
    </para>"""))

def _crossSection(menuitem, mesh, name, cs):
    meshcontext = oof2.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        meshcontext.addCrossSection(name, cs)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    meshcontext.changed()

meshmenu.addItem(OOFMenuItem(
    'CrossSection',
    callback=_crossSection,
    params=[
    whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    parameter.StringParameter('name', tip="Name of the cross section."),
    parameter.RegisteredParameter('cs', meshcrosssection.MeshCrossSection,
                                  tip=parameter.emptyTipString)],
    help="Load a Cross Section. Used internally in Mesh data files.",
    discussion="""<para>
    Load a <link linkend='RegisteredClass:MeshCrossSection'>cross
    section</link>.
    </para>"""
    ))
    

def _assignMat(menuitem, mesh, elements, material):
    meshcontext = oof2.engine.mesh.meshes[mesh]
    meshcontext.reserve()
    meshcontext.begin_writing()
    try:
        matl = materialmanager.getMaterial(material)
        # Get actual Element from element index
        for eIndex in elements:
            elephant = meshcontext.getObject().getElement(eIndex)
            elephant.setExplicitMaterial(matl)
    finally:
        meshcontext.end_writing()
        meshcontext.cancel_reservation()
    meshcontext.changed()

meshmenu.addItem(OOFMenuItem(
    'AssignMaterial',
    callback=_assignMat,
    params=[
    whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                          tip=parameter.emptyTipString),
    parameter.ListOfIntsParameter('elements', tip='Indices of the elements'),
    parameter.StringParameter('material',
                              tip='Material to assign to the elements')],
    help="Load explicitly assigned Materials into a Mesh",
    discussion="""<para>
    Explicitly assign a &material; to &elems; in a saved &mesh;.  See
    <xref linkend='RegisteredClass:AssignMaterial'/>.
    </para>"""
    ))
