# -*- python -*-
# $RCSfile: meshinfo.py,v $
# $Revision: 1.53 $
# $Author: langer $
# $Date: 2005/05/13 18:44:09 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import switchboard
from oof2.SWIG.common import timestamp
from oof2.common import debug
from oof2.common import primitives
from oof2.common import ringbuffer
from oof2.common import toolbox
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump

class MeshQueryContainer:
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context  # mesh context
        self.object = None
        self.position = None
        self.targetname = None

    def set(self, object=None, targetname=None, position=None):
        self.object = object
        self.targetname = targetname
        self.position = position
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp

    def clone(self):
        krusty = MeshQueryContainer(self.context)
        krusty.object = self.object
        krusty.targetname = self.targetname
        krusty.position = self.position
        krusty.timestamp.increment()
        return krusty

    def clearable(self):
        return not not self.object

# Peek functionality for "mesh" is going to be limited to "Node".
# There's a remote chance to include "Element" and "Edge" here.
# However, the code is written in a form that can handle these remote chance.
class MeshPeekContainer(MeshQueryContainer):
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context
        self.objects = {"Node":None}

    def assignObject(self, object, objtype):
        self.objects[objtype] = object
        self.timestamp.increment()

    def removeObject(self, objtype):
        self.objects[objtype] = None
        self.timestamp.increment()

    def reset(self):
        self.objects = {"Node":None}
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp


#############################################################
            
class MeshInfoToolbox(toolbox.Toolbox):
    def __init__(self, gfxwindow):
        toolbox.Toolbox.__init__(self, 'Mesh_Info', gfxwindow)
        self.whoset = ('Mesh',)
        self.querier = None
        self.peeker = None      
        self.records = ringbuffer.RingBuffer(49)

        self.sbcallbacks = [
            switchboard.requestCallback((self.gfxwindow(), "layers changed"),
                                        self.newLayers)]

    def close(self):
        map(switchboard.removeCallback, self.sbcallbacks)

    def makeMenu(self, menu):
        self.menu = menu
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryElement',
            callback=self.queryElem,
            params=[primitives.PointParameter('position', tip='Target point.')],
            help="Query an FE element at a given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/query_mesh_elem.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryNode',
            callback=self.queryNode,
            params=[primitives.PointParameter('position', tip='Target point.')],
            help="Query an FE node at a given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/query_mesh_node.xml')
            ))

    def activate(self):
        meshcontext = apply(self.gfxwindow().topwho, self.whoset)
        if meshcontext is not None:
            if not self.querier:
                self.querier = MeshQueryContainer(meshcontext)
            if not self.peeker:
                self.peeker = MeshPeekContainer(meshcontext)
        else:
            self.querier = None
            self.peeker = None

    def newLayers(self):
        meshcontext = apply(self.gfxwindow().topwho, self.whoset)
        if meshcontext is not None:
            if self.querier and \
               meshcontext is not self.querier.context:
                self.resetRecords()
                self.querier = MeshQueryContainer(meshcontext)
                self.peeker = MeshPeekContainer(meshcontext)
                self.redraw()
            else:
                self.querier = MeshQueryContainer(meshcontext)
                self.peeker = MeshPeekContainer(meshcontext)
        else:
            if self.querier is not None:
                self.querier.set()
            if self.peeker is not None:
                self.peeker.reset()
            self.resetRecords()

    def resetRecords(self):
        self.records.clear()

    def resetPeeker(self):
        if self.peeker is not None:
            self.peeker.reset()

    def queryElem(self, menuitem, position):
        context = apply(self.gfxwindow().topwho, self.whoset)
        if not context:
            return
        femesh = context.getObject()
        skeleton = context.getSkeleton()
        selem = skeleton.enclosingElement(position)
        felem = femesh.getElement(selem.meshindex)
        self.finishQuery(felem, "Element", position)

    def queryNode(self, menuitem, position):
        context = apply(self.gfxwindow().topwho, self.whoset)
        if not context:
            return
        femesh = context.getObject()
        fnode = femesh.closestNode(position.x, position.y)
        self.finishQuery(fnode, "Node", position)

    def finishQuery(self, object, targetname, position):
        self.querier.set(object=object, targetname=targetname,
                         position=position)
        self.records.push(self.querier.clone())
        self.peeker.reset()
        self.querysignal()

    def querysignal(self):
        switchboard.notify((self.gfxwindow(),"query mesh"))
        switchboard.notify((self.gfxwindow(), "meshinfo click"),
                           self.querier.position)
        self.redraw()

    def clearQuerier(self):
        if self.querier:
            self.querier.set()
            self.peeker.reset()
            self.redraw()

    def currentPoint(self):
        try:
            currentquerier = self.records.current()
        except IndexError:
            return
        if currentquerier:
            return currentquerier.position

    def redraw(self):
        switchboard.notify("redraw")

    def prev_able(self):
        return not self.records.atBottom()

    def next_able(self):
        return not self.records.atTop()
    tip = "Get information about a Mesh."
    discussion="""<para>
    Get information about a &mesh;, including &field; values, based on
    mouse input.
    </para>"""

toolbox.registerToolboxClass(MeshInfoToolbox, ordering=3.0)
