# -*- python -*-
# $RCSfile: materialmanager.py,v $
# $Revision: 1.41 $
# $Author: langer $
# $Date: 2005/05/13 18:44:05 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# The MaterialList is a database of all current materials, each
# of which is an aggregate of properties corresponding to entries
# in the propertyTree.

from oof2.engine.propertyregistration import AllProperties
from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import material
from oof2.common.microstructure import microStructures

from oof2.SWIG.common import ooferror
from oof2.common.IO import reporter
from oof2.common import utils
from oof2.common import debug


# "MaterialProps" is a collection of properties which make up
# a material.  It contains the actual material instance, which
# has within it actual property instances, suitable for importation
# into a mesh.

#   Data entries are property-registration entries corresponding
# to nodes in the PropertyManager.
#   This class has additional GUI-related functionality
# in engine/IO/GUI/materialListGUI.py.
class MaterialProps:
    def __init__(self,name,actual=None):
        self.current_prop = None  
        self.name = name
        self.data = {}
        self.actual = actual or material.newCMaterial(name)
    def __repr__(self):
        return "MaterialProps('%s')" % self.name
    
    # Detect name collisions, and check against existing props
    # for category collisions.
    # For starters, "key" is the fully qualified property name,
    # and "value" is the associated PropertyRegistration object.
    # Store a reference to a copy of the property instance.
    def __setitem__(self,key,value):
        try:
            collision = self.data[key]
        except KeyError:
            propcopy = value()
            self.data[key]=value
            self.actual.addProperty(propcopy)
            value.materials[self.name]=(self, propcopy)
            switchboard.notify("prop_added_to_material",self, key, value)
            switchboard.notify("redraw")
        else:
            raise KeyError("Collision in Material Property List, key %s." % \
                           key)

    # Add a propertyregistration entry to the dict without
    # adding the property to the "actual".  Used when loading
    # a material from a script, in which case the material already
    # has the property.
    def add_prop_ref(self, key, value):
        self.data[key]=value
        switchboard.notify("prop_added_to_material", self, key, value)
        
    def __getitem__(self,key):
        return self.data[key]

    # Remove the named property from this material, if it occurs.
    # Property should *not* be current by this time.
    def delete_prop(self,name):
        if self.data.has_key(name):
            reg = self.data[name]
            self.actual.removeProperty(reg.materials[self.name][1])
            switchboard.notify("prop_removed_from_material",
                               self, name, reg)
            switchboard.notify("redraw")
            del reg.materials[self.name]
            del self.data[name]

    def cleanAttributes(self, ms):
        self.actual.cleanAttributes(ms)

    # When a property gets new parameters, you need to update
    # the "actual" entry, but nothing else really changes.
    # "actual" properties are Python-wrapped Property instances.
    def new_params(self, old_prop, new_prop):
        self.actual.removeProperty(old_prop)
        self.actual.addProperty(new_prop)

        
# This class's GUI functionality is also in
# engine/IO/GUI/materalListGUI.py

class MaterialManager:
    def __init__(self):
        # self.current_material = None
        self.materials={}

    # Useful for returning a material to a script.
    def __getitem__(self, key):
        return self.materials[key]
    
    # Delete this property from each material in which it occurs.
    # "Deletion" means it's been deleted from the property manager,
    # and all instances need to go. 
    def delete_prop(self,name):
        for m in self.materials.values():
            m.delete_prop(name)
            
    # Add a property to the named material.
    def add_prop(self, matname, propname):
        prop_reg = AllProperties[propname]
        mat = self[matname]
        if prop_reg: # Can be None for nonparametrizable properties.
            mat[propname]=prop_reg # Triggers GUI callbacks.
        else:
            reporter.report("Nonparametrizable property is not loadable.")

        
    # Remove a property.  "Removal" means remove this particular property
    # from this material -- the property continues to exist in the
    # propertymanager, and possibly in other materials.
    def remove_prop(self, matname, propname):
        mat = self[matname]
        prop = AllProperties[propname]
        if mat and prop:
            mat.delete_prop(propname)
        else:
            # Report a sensible error.
            pass
        

    # Add a material.
    def add(self,name):
        try:
            collision = self.materials[name]
        except KeyError:
            self.materials[name]=MaterialProps(name)
            switchboard.notify("new_material", name, self.materials[name])
            return self.materials[name]
        else:
            raise KeyError("Collision in MaterialManager, key %s." % name)

    # Remove the named material.
    # Menu item checks/reports on current material.
    def delete(self,name):
        doomed = self.materials[name]
        if doomed:
            del self.materials[doomed.name]
            switchboard.notify("remove_material",doomed)
            
        # Ensure that all MaterialAttribute objects which refer to
        # this material are notified of its untimely demise.
        for ms in microStructures.actualMembers():
            doomed.cleanAttributes(ms.getObject())
        switchboard.notify("redraw")
        

    # Material-manager function for loading a material and properties
    # together.  Creates the C material, the MatProps object, and
    # adds it to the materialmanager dictionary, bypassing the
    # functionality used by materialmanager.add, and subsequent
    # add_properties calls.  That set of operations is intended
    # for constructing new materials with no properties.
    def new_material(self, name, *props):
        # Get the C version of the material.
        c_mat = material.newCMaterial(name, *props)

        new_matprop = MaterialProps(name, actual=c_mat)
        self.materials[name] = new_matprop
        switchboard.notify("new_material", name, new_matprop) 

        for p in c_mat.properties():
            reg = p.registration()
            path = AllProperties.data.reverse_dict[reg].path()
            new_matprop.add_prop_ref(path, reg)
            reg.materials[name]=(new_matprop, p)

        return c_mat
    
    # Makes a copy of material 'name', and names it 'newname'.
    def copy(self, name, newname):
        mat = self.materials[name] 
        if mat:
            newmat = self.add(newname) 
            for k in mat.data.keys():
                # add_prop takes care of setting prop.materials.
                self.add_prop(newname, k)
                
        
materialmanager = MaterialManager()
AllProperties.materialmanager = materialmanager

def getMaterialNames():
    return materialmanager.materials.keys()

def getMaterial(name):
    return materialmanager.materials[name].actual

def nMaterials():
    return len(materialmanager.materials)

utils.OOFdefine('getMaterial', getMaterial)
utils.OOFdefine("AllMaterials", materialmanager)

