// -*- C++ -*-
// $RCSfile: piezoelectricity.C,v $
// $Revision: 1.20 $
// $Author: langer $
// $Date: 2005/03/10 22:13:57 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>


#include "common/coord.h"
#include "engine/celectricfield.h"
#include "engine/IO/propertyoutput.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/rank3tensor.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/cstrain.h"
#include "engine/property/orientation/orientation.h"
#include "piezoelectricity.h"


PiezoElectricity::PiezoElectricity(PyObject *reg, 
				   const std::string &nm)
  : Property(nm,reg)
{
  
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
  displacement = dynamic_cast<TwoVectorField*>(Field::getField("Displacement"));
  voltage=dynamic_cast<ScalarField*>(Field::getField("Voltage"));
  total_polarization = dynamic_cast<VectorFlux*>(Flux::getFlux("Total_Polarization"));
}


void PiezoElectricity::precompute() {

}

int PiezoElectricity::K_integration_order(int degree) const {
  /*
    see comment on property.h
  */
  return 2*degree-1; 
}


void PiezoElectricity::fluxmatrix(const FEMesh *mesh,
				  const Element *element,
				  const ElementFuncNodeIterator &nu, 
				  const Flux *flux, FluxData *fluxdata,
				  const MasterPosition &pos) const
{
  const Cijkl modulus = elasticity->cijkl(mesh, element, pos);
  const Rank3Tensor eijk = modulus*_dijkLab;
  //shape functions and their derivatives
  double sf = nu.shapefunction(pos);
  double dsf0 = nu.dshapefunction(0, pos);
  double dsf1 = nu.dshapefunction(1, pos);
  

  //contributions to the stress flux
  if(*flux == *stress_flux)
    for(SymTensorIterator ij; !ij.end(); ++ij) {
      double &mtx_el = fluxdata->matrix_element(mesh, ij, voltage, nu); //returns flux matrix
      // Strain tensor components.
      for(SymTensorIterator kl; !kl.end(); ++kl) {
	double ca = modulus(ij, kl)*(_dijkLab(0,kl)*dsf0 + _dijkLab(1,kl)*dsf1);
	if(!voltage->in_plane(mesh)) 
	  {
	    ca += modulus(ij, kl)*_dijkLab(2,kl)*sf;
	  }
	if(kl.diagonal()) {
	  mtx_el += ca;
	}
	else {
	  mtx_el += 2.0*ca;
	}
      }
    }
    
  
  //contributions to total_polarization flux
    
  // Loop over flux components.  Loop over all components, even if
  // the flux is in-plane, because the out-of-plane components of
  // the flux matrix are used to construct the constraint equation.
  if(*flux == *total_polarization)
    for(VectorFieldIterator i; !i.end(); ++i) {
      // in-plane voltage gradient contributions
      fluxdata->matrix_element(mesh, i, displacement, nu) +=
	eijk(i.integer(), 0, 0) * dsf0 + 
	eijk(i.integer(), 1, 0) * dsf1 + 
	eijk(i.integer(), 0, 1) * dsf0 + 
	eijk(i.integer(), 1, 1) * dsf1 +
	eijk(i.integer(), 0, 2) * dsf0 + 
	eijk(i.integer(), 1, 2) * dsf1;
      // out-of-plane strain contribution
      if(!displacement->in_plane(mesh)) {
	if (i.integer()==2)
	fluxdata->matrix_element(mesh, i, displacement->out_of_plane(), nu) += 
	  eijk(2, i.integer(), 0) * sf +
	  eijk(2, i.integer(), 1) * sf +
	  eijk(2, i.integer(), 2) * sf;
      } //out of plane contributions start here
      
    }   // Loop over all tensor components.
    
  
 
  
}



void PiezoElectricity::output(const FEMesh *mesh,
				  const Element *element,
				  const PropertyOutput *output,
				  const MasterPosition &pos,
				  OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    // The parameter is a Python StrainType instance.  Extract its name.
    acquirePyLock();
    std::string stype;
    try {
      PyObject *str = PyObject_CallMethod(output->parameters()[0], "name", 0);
      stype = PyString_AsString(str);
      Py_XDECREF(str);
    }
    catch(...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    VECTOR_D E_field(3,0);
    findElectricField(mesh, element, pos, E_field);
    if(stype == "Piezoelectric")
      *sdata += _dijkLab*E_field;
    else if(stype == "Elastic")
      *sdata -= _dijkLab*E_field;
  }//strain output ends here
  if(outputname == "Energy") {
    // The parameter is a Python Enum instance.  Extract its name.
    std::string etype;
    acquirePyLock();
    try {
      PyObject *str = PyObject_GetAttrString(output->parameters()[0], "name");
      etype = PyString_AsString(str);
      Py_XDECREF(str);
    }
    catch(...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
    //Energies
    if(etype == "Total" || etype == "Elastic" || etype == "Electric") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      SymmMatrix piezoelectricStrain(3);
      const Cijkl modulus = elasticity->cijkl(mesh, element, pos); //returns lab reference system stiffness
      
      VECTOR_D E_field(3,0);
      findElectricField(mesh, element, pos, E_field);
      piezoelectricStrain = _dijkLab*E_field;
      SymmMatrix electricstress(modulus*piezoelectricStrain);
      SymmMatrix strain(3);
      findGeometricStrain(mesh, element, pos, strain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += electricstress(i,i)*strain(i,i);
	int j = (i+1)%3;
	e += 2*electricstress(i,j)*strain(i,j);
      }
      if(etype == "Total")
	*edata += -e;      
      else if (etype == "Elastic" || etype == "Electric")
	*edata += -0.5*e; 
    } //dfferent types of energies end here
    
  }//energy output ends here
}


int PiezoElectricity::is_symmetric(const FEMesh* mesh) const {
  //std::vector<Equation*> &eqns = Equation::alleqns();
  return 1;
}

IsotropicPiezoElectricity::IsotropicPiezoElectricity(PyObject *registry,
						     const std::string &name, 
						     double d)
  : PiezoElectricity(registry, name),
    _dijkValue(d)
{

}


void IsotropicPiezoElectricity::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  
}

void IsotropicPiezoElectricity::precompute() {
  PiezoElectricity::precompute();
  _dijkLab(0,0,0) = _dijkLab(1,1,1)
    = _dijkLab(2,2,2) = _dijkValue;
}

AnisotropicPiezoElectricity::AnisotropicPiezoElectricity(PyObject *registry, 
							 PyObject *self,
							 const std::string &nm, 
							 Rank3Tensor *dijkTensor)
  :  _dijkValue(*dijkTensor),
     PiezoElectricity(registry, nm),
     PythonNative<Property>(self) 
{

}
void AnisotropicPiezoElectricity::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationProp*>
    (mat->fetchProperty("Orientation"));
}

void AnisotropicPiezoElectricity::precompute()
{
  _dijkLab = _dijkValue.transform(orientation->eulerangle());
}
