# -*- python -*-
# $RCSfile: skeletonelement.py,v $
# $Revision: 1.131 $
# $Author: langer $
# $Date: 2005/05/31 14:42:55 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.SWIG.common import coord
from oof2.SWIG.engine import ooferror2
from oof2.SWIG.engine import cskeleton
from oof2.SWIG.engine import material
from oof2.engine import skeletonselectable
import oof2.SWIG.common.timestamp
from oof2.common import enum
from oof2.common import debug
from oof2.common import object_id
from oof2.common import primitives
from oof2.common import utils
import math

class CachedValue:
    def __init__(self, value=None, timestamp=None):
        self.__dict__['value'] = value
        self.__dict__['savedValue'] = None
        if timestamp is None:
            self.__dict__['timestamp'] = oof2.SWIG.common.timestamp.TimeStamp()
        else:
            self.__dict__['timestamp'] = timestamp.clone()
        if value is None:
            self.timestamp.backdate()
        self.savedTime = oof2.SWIG.common.timestamp.TimeStamp()
        self.savedTime.backdate()
    def __setattr__(self, attr, value):
        if attr == 'value':
            self.__dict__['savedValue'] = self.value
            # Save the timestamp associated with the *old* value, and
            # increment the timestamp associated with the *new* value.
            # TimeStamp.cloneAndIncrement clones a TimeStamp,
            # increments the original, and returns the clone.
            self.__dict__['savedTime'] = self.timestamp.cloneAndIncrement()
        self.__dict__[attr] = value
    def revert(self):
        # Unconditionally go back to the previous value
        self.__dict__['value'] = self.savedValue
        self.__dict__['timestamp'] = self.savedTime
    def clone(self):
        return CachedValue(value=self.value, timestamp=self.timestamp)
    def __repr__(self):
        return "CachedValue(%s)" % `self.value`

#########################################

class ElementShapeType(enum.EnumClass(('triangle', 'triangular elements'),
                                      ('quad', 'quadrilateral elements'))):
    tip = "Shapes of Skeleton Elements."
    discussion = """<para>
    <classname>ElementShapeType</classname> objects are used in a few
    cases where it's necessary to distinguish between &elem; shapes in
    the &oof2; commands.
    </para>"""
    
#########################################

#TODO: Merge HomogeneityCacheData and Homogeneity classes.

class HomogeneityCacheData:
    def __init__(self, homogeneity=None, dominantpixel=None, homogEnergy=0.0):
        self.homogeneity = homogeneity
        self.dominantpixel = dominantpixel
        self.homogEnergy = homogEnergy
    def clone(self):
        return HomogeneityCacheData(self.homogeneity, self.dominantpixel,
                                    self.homogEnergy)
    def __repr__(self):
        return "(%g,%d,%g)" % (self.homogeneity, self.dominantpixel,
                               self.homogEnergy)

#########################################

# The SkeletonElementBase is the base class for both the
# SkeletonElement and the ProvisionalElement.  It can compute
# geometrical stuff like homogeneity and shape energy, but knows
# nothing about selectability, segments, or neighbors.

class SkeletonElementBase:
    def __init__(self, nodes):
        self.nodes = nodes
        self.cachedHomogData = CachedValue()
        
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.nodes)

    def homogeneity(self, skeleton):
        self.findHomogeneityAndDominantPixel(skeleton)
        return self.cachedHomogData.value.homogeneity
    def dominantPixel(self, skeleton):
        self.findHomogeneityAndDominantPixel(skeleton)
        return self.cachedHomogData.value.dominantpixel
    def material(self, skeleton):
        self.findHomogeneityAndDominantPixel(skeleton)
        dominantpixel = self.cachedHomogData.value.dominantpixel
        if dominantpixel is not None:
            return material.getMaterialFromCategory(skeleton.MS, dominantpixel)

    def findHomogeneityAndDominantPixel(self, skeleton):
        # Is the already calculated homogeneity up-to-date?  It is
        # up-to-date if no node has moved since it was last computed,
        # and if the microstructure hasn't changed.  cachedHomogEnergy
        # changes every time cachedHomogeneity changes, so it doesn't
        # need to be separately tracked.
        homogTimeStamp = self.cachedHomogData.timestamp
        if homogTimeStamp > skeleton.MS.getTimeStamp():
            for node in self.nodes:
                if node.nodemoved > homogTimeStamp:
                    break;              # not up-to-date
            else:                       # up-to-date
                return

        if self.illegal():
            self.cachedHomogData.value = HomogeneityCacheData()
            return

        microstructure = skeleton.MS

        homog = self.c_homogeneity(microstructure); # C Homogeneity struct
        self.cachedHomogData.value = HomogeneityCacheData(
            homog.homogeneity, homog.category, self.homogeneityEnergy(homog))

    def copyHomogeneity(self, other):
        # This doesn't just use CachedValue.clone, because then the
        # other element and this one would share a reference to the
        # same HomogeneityCacheData object.
        if other.cachedHomogData.value is not None:
            self.cachedHomogData.value = other.cachedHomogData.value.clone()
            self.cachedHomogData.timestamp = other.cachedHomogData.timestamp.clone()

    def revert(self):
        # Revert cached calculations UNCONDITIONALLY.
        # This is used to restore the state after a failed node
        # motion.
        self.cachedHomogData.revert()
        
    def underlyingPixels(self, microstructure):
        # Returns a list of all pixels that overlap with this element.
        dx, dy = microstructure.sizeOfPixels()
        nx, ny = microstructure.sizeInPixels()
        return self.underlying_pixels(microstructure) # C++

    def transitionPoint(self, skeleton, edgeno):
        ok, point = \
            cskeleton.CSkeletonElementPtr.transitionPoint(self, skeleton.MS,
                                                          edgeno)
        if ok:
            return point
    
    def energyHomogeneity(self, skeleton):
        self.findHomogeneityAndDominantPixel(skeleton)
        return self.cachedHomogData.value.homogEnergy

    def energyTotal(self, skeleton, alpha):
        if alpha == 0:
            return self.energyShape()
        if alpha == 1:
            return self.energyHomogeneity(skeleton)
        return alpha*self.energyHomogeneity(skeleton) + \
               (1.-alpha)*self.energyShape()

    def setHomogeneityValues(self, homogeneity):
        self.cachedHomogData.value = HomogeneityCacheData(
            homogeneity.homogeneity,
            homogeneity.category,
            self.homogeneityEnergy(homogeneity))
        
##########################
        
class SkeletonElement(SkeletonElementBase,
                      skeletonselectable.SkeletonSelectable):
    count = 0
    def __init__(self, nodes):
        SkeletonElementBase.__init__(self, nodes)
        skeletonselectable.SkeletonSelectable.__init__(self)

        self.index = SkeletonElement.count
        SkeletonElement.count += 1

        # Although the CSkeletonElement keeps a list of
        # CSkeletonNodes, the extra information in the Python
        # SkeletonNodes isn't available unless we keep a list of them
        # here, as well.  It's possible that we could move all the
        # extra info into the CSkeletonNode class and swig it.  If so,
        # we'd have to be careful when using utils.canonicalorder() on
        # nodes, since the CSkeletonNodePtrs returned by swig won't
        # have consistent ids.
        self.nodes = nodes
        
        for node in nodes:
            node.addElement(self)
            
        # When a real mesh is made from the skeleton in which this
        # element lives, self.meshindex gets assigned the index of
        # that element.  This index is the same for all the real meshes.
        self.meshindex = None

        self.ID = object_id.ObjectID()
        
    def repr_position(self):
        return self.center()

    def getIndex(self):
        return self.index

    def type(self):
        return self.shapetype

    def destroy(self, skeleton):
        self.disconnect()
        nnodes = self.nnodes()
        lastnode = self.nodes[-1]
        for node in self.nodes:
            skeleton.findSegment(lastnode, node).removeElement(skeleton, self)
            node.removeElement(skeleton, self)
            lastnode = node

    # def getFEelement(self):
    #     return self.element

    def getSegments(self, skeleton):
        segments = []
        lastnode = self.nodes[-1]
        for node in self.nodes:
            segments.append(skeleton.findSegment(node, lastnode))
            lastnode = node
        return segments

    def active(self, skeleton):
        for node in self.nodes:
            if node.active(skeleton):
                return 1
        return 0

    def new_child(self):
        raise ooferror.ErrPyProgrammingError(
            "Attempt to clone element parent class.")

    def segment_iterator(self, skel):
        segments = []
        for i in range(self.nnodes()):
            n0 = self.nodes[i]
            n1 = self.nodes[(i+1)%self.nnodes()]
            segments.append(skel.findSegment(n0, n1))
        return segments

    def edgeNeighbors(self, skeleton, loopRange=0):
        # Search for edge-sharing neighborhood for a given element and return
        # a list of them.
        nnodes = self.nnodes()
        edgeNeighborList = [None]*nnodes
        for i in range(nnodes):
            segment = skeleton.findSegment(self.nodes[i],
                                           self.nodes[(i+1)%nnodes])
            if segment is not None:
                edgeNeighborList[i] = segment.getOtherElement(self)

        return edgeNeighborList

    def getEdgeLengthsList(self):
        list = []
        for i in range(self.nnodes()):
            list.append(self.edgeLength(i))
        return list

    def getShortestEdge(self):
        list = self.getEdgeLengthsList()
        minEdge = list[0]
        index = 0
        for i in range(1, self.nnodes()):
            if list[i] < minEdge:
                minEdge = list[i]
                index = i
        return index
    
    # Test code for getting a real angle at a given node index
    # Probably will move to C++
    def getRealAngle(self, i):
        # Test code for getting real angle(degree) of each corner
        # including illegal ones - returned as negative angle
        C = self.nodes[(i-1)%self.nnodes()].position()
        B = self.nodes[i].position()
        A = self.nodes[(i+1)%self.nnodes()].position()
        dot = (A.x-B.x)*(C.x-B.x) + (A.y-B.y)*(C.y-B.y)
        cross = (A.x-B.x)*(C.y-B.y) - (A.y-B.y)*(C.x-B.x)
        return math.atan2(cross, dot)*180.0/math.pi

    def getAnglesList(self):
        list = []
        for i in range(self.nnodes()):
            list.append(self.cosCornerAngle(i))
        return list
    
    def getBiggestAngle(self):
        list = self.getAnglesList()
        maxAngle = list[0]  # cosine of the angle
        index = 0
        for i in range(1, self.nnodes()):
            if list[i] < maxAngle:
                maxAngle = list[i]
                index = i
        return index

    def getSister(self, skeleton, node0, node1):
        # 4 _________ 3
        #   \       /
        #    \  A  /    If an edge 1-2 of the element A needs to be modified
        #     \   /     (for example, collapsing), it is important to update
        #    1 \_/ 2    its edge-sharing neighbor element B.
        #      / \      This function returns a sister element of "self" which
        #     /   \     shares an edge (node1-node2).
        #    /  B  \
        #   ---------

        segment = skeleton.findSegment(node0, node1)
        elements = segment.getElements()
##        if len(elements) > 2:
##            raise ooferror.ErrPyProgrammingError("Too many sisters!")
        for e in elements:
            if e is not self:
                return e
        return None  # edge boundary

    #########################

##    def realelement(self, skeleton, mesh, index, fe_node,
##                    curnodeindex, edgedict, elemdict, materialfunc):
    def realelement(self, skeleton, mesh, index, fe_node,
                    curnodeindex, elemdict, materialfunc):
        # Create a real element corresponding to this skeleton
        # element. The elemdict argument is a dictionary (keyed by the
        # number of sides of the element) of MasterElement objects,
        # which contain the information needed to construct the real
        # elements.  /*The edgedict argument is a dictionary of edges --
        # each time nodes are added to an edge of the mesh, they are
        # stored here.  The dictionary keys are tuples of nodes
        # (defining an edge) and the entries are lists of nodes on
        # that edge.*/  "index" is the SkeletonElement's position in the
        # Skeleton's list, and "fe_node" is a dictionary of real nodes
        # in the FEMesh, indexed by their corresponding SkeletonNode
        # objects.  The lists give the nodes in the order in which
        # they were added to the element.
        
        # Be safe with indices.
        if self.meshindex is None: # Zero is nontrivial index.
            self.meshindex = index
        else:
            if index != self.meshindex:
                raise ooferror2.ErrPyProgrammingError(
                    "Index mismatch in element construction.")
                
        nnewnodes = 0
        ncn = len(self.nodes) # Corner nodes.

        # elemdict is a dictionary of MasterElements, keyed by number
        # of sides.
        elementtype = elemdict[self.nnodes()]
        nodes = []                      # real nodes for this element

        for i in range(len(self.nodes)):  # i.e. for each edge...
            c0 = self.nodes[i]
            nodes.append(fe_node[c0])     # Node nodes already exist.
            c1 = self.nodes[(i+1)%ncn]
            cset = utils.canonicalorder(c0, c1)

##            # Look up this edge in the dictionary.  If it's there,
##            # then nodes have been created on the edge already, and we
##            # should reuse them.
##            try:
##                xtranodes = edgedict[cset]
##                # The nodes were created by the neighboring element.
##                # Since elements traverse their edges counterclockwise
##                # when creating nodes, the preexisting nodes are in
##                # the wrong order for the current element.
##                xtranodes.reverse()
##                # Remove cset from the dictionary.  We're done with it.
##                del edgedict[cset]
##            except KeyError:
##                # The edge wasn't in the dictionary.  It's a new edge.
##                xtranodes  = []
##                # Loop over all protonodes on the current
##                # edge of the new element
##                for newproto in elementtype.protodic[i]:
##                    masterxy = primitives.Point(newproto.mastercoord()[0],
##                                           newproto.mastercoord()[1])
##                    realxy = self.frommaster(masterxy, 0)
##                    newnode = _makenewnode(mesh, newproto,
##                                           coord.Coord(realxy.x, realxy.y))
##                    nnewnodes = nnewnodes + 1
##                    xtranodes.append(newnode)
##                edgedict[cset] = xtranodes
##            nodes = nodes + xtranodes

            # Look up this edge in the dictionary.  If it's there,
            # then nodes have been created on the edge already, and we
            # should reuse them.

            try:
                xtranodes = mesh.getEdgeNodes(cset)
                # The nodes were created by the neighboring element.
                # Since elements traverse their edges counterclockwise
                # when creating nodes, the preexisting nodes are in
                # the wrong order for the current element.
                xtranodes.reverse()
            except KeyError:
                # The edge wasn't in the dictionary.  It's a new edge.
                xtranodes  = []
                # Loop over all protonodes on the current
                # edge of the new element
                for newproto in elementtype.protodic[i]:
                    masterxy = primitives.Point(newproto.mastercoord()[0],
                                           newproto.mastercoord()[1])
                    realxy = self.frommaster(masterxy, 0)
                    newnode = _makenewnode(mesh, newproto,
                                           coord.Coord(realxy.x, realxy.y))
                    nnewnodes = nnewnodes + 1
                    xtranodes.append(newnode)
                mesh.addEdgeNodes(cset, xtranodes)
            nodes = nodes + xtranodes

        # Interior nodes at the end.
        for newproto in elementtype.protodic['interior']:
            masterxy = primitives.Point(newproto.mastercoord()[0],
                                        newproto.mastercoord()[1])
            realxy = self.frommaster(masterxy, 0)
            newnode = _makenewnode(mesh, newproto,
                                   coord.Coord(realxy.x, realxy.y))
            nnewnodes = nnewnodes + 1
            nodes.append(newnode)
            mesh.addInternalNodes(self, newnode)

        # Having constructed the list of nodes, call the real
        # element's constructor.
        ## realel = elementtype.build(self, self.material(skeleton), nodes)
        realel = elementtype.build(self, materialfunc(self, skeleton), nodes)
        # Long-lost cousin of Kal-El and Jor-El.
        
        mesh.AddElement(realel)               # Add to mesh.
        # Tell the element about its exterior edges.
        for edge in self.exterior_edges:
            realel.set_exterior(fe_node[edge[0]], fe_node[edge[1]])

        return nnewnodes

    def realmaterial(self, skeleton):
        return self.material(skeleton)

#####################################################

class ProvisionalElement(SkeletonElementBase):
    def __init__(self, nodes, parents):
        SkeletonElementBase.__init__(self, nodes)
        self.parents = parents
    def accept(self, skeleton):
        element = skeleton.newElement(self.nodes, parents=self.parents)
        element.copyHomogeneity(self)
        return element

#####################################################    

class SkeletonQuad(SkeletonElement, cskeleton.CSkeletonQuad):
    # A four-sided ghost element with master space (-1,-1) to (1,1).
    def __init__(self,nodes):
        cskeleton.CSkeletonQuad.__init__(self, *nodes)
        SkeletonElement.__init__(self,nodes)

    shapetype = ElementShapeType('quad')

##    def __repr__(self):
##        return "SkeletonQuad"

    def new_child(self):
        node_children = [ x.getChildren()[-1] for x in self.nodes ]
        new = SkeletonQuad(node_children)
        new.copyHomogeneity(self)
        return new

    def aspectRatio(self):
        #         ________d(max)_________
        #        |                      /
        # a(min) |                     / c
        #        |____________________/
        #                  b 
        #
        # aspect ratio : 1. Find a max ratio of two consecutive segments
        #                   => d/a
        #                2. Get an opposite segment of "a". => c
        #                3. A.R. => max(c,d)/min(c,d)
        #
        # The reason why not using d/a as A.R. is that if a quad. has one
        # short segement and three similarly long segments, d/a may give
        # a wrong impression about this quad.

        lengths = [self.edgeLength(i) for i in range(4)]
        max_ratio = 1.0
        long_side = None
        short_side = None
        for i in range(4):
            if lengths[i] >= lengths[(i+1)%4]:
                long = i
                short = (i+1)%4
            else:
                long = (i+1)%4
                short = i
            ratio = lengths[long]/lengths[short]
            if ratio >= max_ratio:
                max_ratio = ratio
                long_side = long
                short_side = short
        opposite = (short_side+2)%4
        return max(lengths[opposite], lengths[long_side])/\
               min(lengths[opposite], lengths[long_side])

##    def getLongSegments(self, skeleton):
##        lengths = [self.edgeLength(i) for i in range(4)]
##        long = lengths.index(max(lengths))
##        seg1 = skeleton.findSegment(self.nodes[long], self.nodes[(long+1)%4])
##        seg2 = skeleton.findSegment(self.nodes[(long+2)%4],
##                                    self.nodes[(long+3)%4])
##        return seg1, seg2

    def getAspectRatioSegments(self, threshold, skeleton):
        ## Return segments that should be refined by CheckAspectRatio.
        ## The long edges of quads can be refined if there are two
        ## long edges and two short ones, and the long ones are
        ## opposite to each other.  We check the ratio of the second
        ## longest to the second shortest, because this rules out
        ## cases in which one edge is much longer than the other
        ## three.
        segs = []
        lengths = map(self.edgeLength, (0,1,2,3))
        sortlengths = lengths[:]
        sortlengths.sort()
        if sortlengths[2] > sortlengths[1]*threshold:
            # find longest edges
            l0 = sortlengths[2]
            l1 = sortlengths[3]
            i0 = i1 = None
            for i in (0,1,2,3):
                leng = lengths[i]
                if leng == l0:
                    i0 = i
                if leng == l1:
                    i1 = i
            idiff = i0 - i1
            if idiff == 2 or idiff == -2:
                segs.append(skeleton.findSegment(self.nodes[i0],
                                                self.nodes[(i0+1)%4]))
                segs.append(skeleton.findSegment(self.nodes[i1],
                                                 self.nodes[(i1+1)%4]))
        return segs

class ProvisionalQuad(ProvisionalElement, cskeleton.CSkeletonQuad):
    def __init__(self, nodes, parents):
        cskeleton.CSkeletonQuad.__init__(self, *nodes)
        ProvisionalElement.__init__(self, nodes, parents)

##################################
        
class SkeletonTriangle(SkeletonElement, cskeleton.CSkeletonTriangle):
    # A three-sided ghost element with master space (1,0) -> (0,1) -> (0,0)
    def __init__(self,nodes):
        cskeleton.CSkeletonTriangle.__init__(self, *nodes)
        SkeletonElement.__init__(self,nodes)

    shapetype = ElementShapeType('triangle')

    def new_child(self):
        node_children = [ x.getChildren()[-1] for x in self.nodes ]
        new = SkeletonTriangle(node_children)
        new.copyHomogeneity(self)
        return new

    def aspectRatio(self):
        lengths = [self.edgeLength(i) for i in range(3)]
        lengths.sort()
        ## This returns the ratio of the *middle* length to the
        ## *shortest* length, which might seem odd. It's a more
        ## conservative definition than using lengths[2]/lengths[0].
        ## The reason for using it is that aspectRatio is used to mark
        ## elements for refining, and not much is gained by refining
        ## high aspect triangles unless they're more or less
        ## isosceles, with only one small angle.
        return lengths[1]/lengths[0]

##    def getLongSegments(self, skeleton):
##        lengths = [self.edgeLength(i) for i in range(3)]
##        short = lengths.index(min(lengths))
##        seg1 = skeleton.findSegment(self.nodes[(short+1)%3],
##                                    self.nodes[(short+2)%3])
##        seg2 = skeleton.findSegment(self.nodes[(short+2)%3],
##                                    self.nodes[(short+3)%3])
##        return seg1, seg2

    def getAspectRatioSegments(self, threshold, skeleton):
        ## return segments that should be refined by CheckAspectRatio
        return []

##    def __repr__(self):
##        return "SkeletonTriangle"

class ProvisionalTriangle(ProvisionalElement, cskeleton.CSkeletonTriangle):
    def __init__(self, nodes, parents):
        cskeleton.CSkeletonTriangle.__init__(self, *nodes)
        ProvisionalElement.__init__(self, nodes, parents)

###########################

def getProvisionalElement(nodes, parents):
    if len(nodes) == 3:
        return ProvisionalTriangle(nodes, parents)
    if len(nodes) == 4:
        return ProvisionalQuad(nodes, parents)

###########################

def _makenewnode(mesh, protonode, coord):
    """Make a new node in mesh at coord, given a protonode."""
    if protonode.func():
        return mesh.newFuncNode(coord)
    return mesh.newMapNode(coord)

############################

